"""This is a module that provides additional jinja filters to help with theme development

Functions
---------
inject_filters() -> jinja2.Environment:
    Takes in a jinja environment and injects the filters from this module + functions fom the extra_filters parameter

split_to_sublists() -> List[list]:
    Takes a list and splits it into sublists of size n

get_image_path() -> str:
    Takes in the path to an image and returns it in usable format to use in img tags as src attribute

get_filename_without_extension() -> str:
    Takes in path and returns filename without extension

pretty_datetime() -> str:
    A utility function for pretty printing dates provided for jobs/getting a degree/volunteering etc

pretty_defaultdict() -> str:
    Returns a prettyprinted form of a defaultdict
"""
# Standard library dependencies
import logging
from pprint import pformat                     # Used to help pretty print dictionaries
from inspect import stack, currentframe        # Used to get error messages from jinja templates
from typing import Callable, DefaultDict, List # Used to typehint accurately for inline documentation


# Third Party Dependencies
import jinja2           # Used mostly to typehint and allow for autocompletion in the file


def inject_filters(env:jinja2.Environment, extra_filters:List[Callable] = None) -> jinja2.Environment:
    """Takes in a jinja environment and injects the filters from this module + functions fom the extra_filters parameter

    Parameters
    ----------
    env : jinja2.Environment
        The existing environment you want to inject the filters into

    extra_filters : List[Callable], optional
        An optional set of method objects containing additional filter functions you want to use

    Returns
    -------
    jinja2.Environment
        The input environment with the filters injected
    """
    if extra_filters is None:
        extra_filters = []
    logging.debug("[ezcv inject_filters()]: Beggining to inject filters")
    filters = [split_to_sublists, get_image_path, get_filename_without_extension, pretty_datetime, pretty_defaultdict]

    if extra_filters:
        logging.debug("[ezcv inject_filters()]: Extra filters found")
        for current_filter in extra_filters:
            filters.append(current_filter)

    for current_filter in filters:
        env.filters[current_filter.__name__] = current_filter

    return env


def split_to_sublists(initial_list:list, n:int, strict:bool=True) -> List[list]:
    """Takes a list and splits it into sublists of size n

    Parameters
    ----------
    initial_list : list
        The initial list to split into sublists

    n : int
        The size of each sublist

    strict: bool
        Whether to force an error if the length of the initial list is not divisible by n (split into even groups), default True

    Returns
    -------
    List[list]
        A list of lists of size n (unless strict is False, then the last list may be > n)

    Examples
    --------

    ### Split gallery images into sublists of 3

    #### JINJA USAGE
    ```jinja2
    {% if gallery|length % 3 == 0 %}
    {% for sublist in gallery|split_to_sublists(3) %}
        <div class="row">

        <div class="col-md-4">
            <img src="{{ sublist.0[0]['file_path'] }}" alt="{{ sublist.0[0]['file_path'].split()[-1] }}">
        </div>

        <div class="col-md-4">
            <img src="{{ sublist.1[0]['file_path'] }}" alt="{{ sublist.1[0]['file_path'].split()[-1]}}">
        </div>

        <div class="col-md-4">
            <img src="{{ sublist.2[0]['file_path'] }}" alt="{{ sublist.2[0]['file_path'].split()[-1] }}">
        </div>

        </div>
    {% endfor %}
    {% endif }
    ```

    The above jinja is roughly equivalent to something like this in pure python:

    ```python
    gallery = ["image 1" , "image 2", "image 3", "image 4" , "image 5", "image 6"]

    if len(images) % 3 == 0:
        for sublist in split_to_sublists(gallery, 3): # Returns [["image 1" , "image 2", "image 3"], ["image 4" , "image 5", "image 6"]]
        ... # Do stuff with each sublist
    ```
    """
    logging.debug(f"[ezcv split_to_sublists({initial_list}, {n}, {strict})]: Beggining filter function")
    if strict and not len(initial_list) % n == 0:
            raise ValueError(f"\033[;31m Provided list was not of correct size: \n\tList: {initial_list}\n\tSegment size {n} \033[0m")

    result = []

    for i in range(0, len(initial_list), n): # Create sublists up to size n 
        result.append( initial_list[i:i + n])

    return result

def get_image_path(path:str) -> str:
    """Takes in the path to an image and returns it in usable format to use in img tags as src attribute

    Parameters
    ----------
    path : str
        The raw image path from metadata

    Returns
    -------
    str
        The string corresponding to the correct path to use in img tags src attributes

    Examples
    --------
    ### Passing in an image path from a project in the projects section:

    #### JINJA USAGE
    ```jinja2
    {% for project in projects %}
        {% if project[0]["image"] %}
            <img src="{{ project[0]['image'] | get_image_path }}" alt="{{ project[0]['image'] | get_filename_without_extension }}" />
        {% endif %}
    {% endfor %}
    ```

    The above jinja is roughly equivalent to something like this in pure python:

    ```python
    project = [{"image": "image.jpg"}, ["other stuff"]]

    if project[0]["image"]:
        print(get_image_path(project[0]['image'])) # Prints /images/image.jpg which is a usable path

    project = [{"image": "https://example.com/img/image.jpg"}, ["other stuff"]]

    if project[0]["image"]:
        print(get_image_path(project[0]['image'])) # Prints https://example.com/img/image.jpg which is a usable path
    ```
    """
    logging.debug(f"[ezcv get_image_path({path})]: Beggining to find image path")
    try:
        if path.startswith("http"):
            return path

        elif path.startswith("images"):
            return f"{path}"

        else:
            return f"images/{path}"
    except AttributeError:
        for frameInfo in stack(): # Get the frame for the error raised
            if frameInfo.frame.f_globals.get("__jinja_template__") is not None: # Find the jinja template namespace if it exists
                template = frameInfo.frame.f_globals.get("__jinja_template__")
                break
        if not path: # If the image path is False (usually because a required image wasn't provided)
            raise ValueError(f"\033[;31m No path provided for a required image in {template.filename} #line {template.get_corresponding_lineno(currentframe().f_back.f_lineno)} \nCheck your themes documentation for details on which images are required (likely in config.yml): https://ezcv.readthedocs.io/en/latest/included-themes/ \033[0m")
        else: # If it's just an invalid image path
            raise ValueError("\033[;31m Could not get image path: {path}\n Error occured on \n{template.filename} #line {template.get_corresponding_lineno(currentframe().f_back.f_lineno)} \nCheck documentation on image management for details https://ezcv.readthedocs.io/en/latest/usage/#image-management\033[0m")

def get_filename_without_extension(path:str) -> str:
    """Takes in path and returns filename without extension

    Parameters
    ----------
    path : str
        The original path to file

    Returns
    -------
    str
        Then name without the extension on the end

    Examples
    --------
    ### Take in an image path and return the filename without extension to use for alt tag

    #### JINJA USAGE
    ```jinja2
    {% for project in projects %}
        {% if project[0]["image"] %}
            <img src="{{ project[0]['image'] | get_image_path }}" alt="{{ project[0]['image'] | get_filename_without_extension }}" />
        {% endif %}
    {% endfor %}
    ```

    The above jinja is roughly equivalent to something like this in pure python:

    ```python
    project = [{"image": "/path/to/John Doe.jpg"}, ["other stuff"]]

    if project[0]["image"]:
        print(get_filename_without_extension(project[0]['image'])) # Prints "John Doe"
    ```
    """
    logging.debug(f"[ezcv get_filename_without_extension({path})]: Beggining to find filename without path")
    return str(path.split("/")[-1].split(".")[0])

def pretty_datetime(month_started:str, year_started:str, month_ended:str, year_ended:str, current:bool) -> str:
    """A utility function for pretty printing dates provided for jobs/getting a degree/volunteering etc

    Parameters
    ----------
    month_started : str
        The month started i.e. October

    year_started : str
        The year started i.e. 2013

    month_ended : str
        The month ended i.e. December

    year_ended : str
        The year ended i.e. 2017

    current : bool
        A boolean describing if this is somewhere you are currently working/studying/volunteering at

    Returns
    -------
    str
        A pretty string of the date i.e. 'October 2013 - December 2017'

    Examples
    --------
    ### Printing the date details of a degree in the `education` section:

    #### JINJA USAGE
    ```jinja2
    {% for experience in education %}
        {{ experience[0]["month_started"] | pretty_datetime(experience[0]["year_started"], experience[0]["month_ended"], experience[0]["year_ended"], experience[0]["current"]) }}
    {%endfor%}
    ```

    The above jinja is roughly equivalent to something like this in pure python:

    ```python

    month_started = "October"
    year_started = "2013"

    month_ended = "December"
    year_ended = "2017 

    current = False

    print(pretty_datetime(month_started, year_started, month_ended, year_ended, current)) # October 2013 - December 2017
    ```
    """
    logging.debug(f"[ezcv pretty_datetime({month_started}, {year_started}, {month_ended}, {year_ended}, {current}))]: Pretty printing datetime")
    
    if month_started or year_started:
        if month_started and year_started:
            beginning = f"{month_started} {year_started}"
        elif month_started:
            beginning = f"{month_started}"
        elif year_started:
            beginning = f"{year_started}"
        sep = " - "
    else:
        sep = ""

    if current:
        end = "Present"
    
    elif month_ended or year_ended:
        if month_ended and year_ended:
            end = f"{month_ended} {year_ended}"
        elif month_ended:
            end = f"{month_ended}"
        elif year_ended:
            end = f"{year_ended}"
    else:
        end = ""
    logging.debug(f"[ezcv pretty_datetime({month_started}, {year_started}, {month_ended}, {year_ended}, {current}))]: result = {beginning}{sep}{end}")
    return f"{beginning}{sep}{end}"


def pretty_defaultdict(ugly_dict:DefaultDict) -> str:
    """Returns a prettyprinted form of a defaultdict

    Parameters
    ----------
    ugly_dict : DefaultDict
        A defaultdictionary to pretty print

    Notes
    -----
    Needs to be used with the safe filter to work properly

    Returns
    -------
    str
        The 

    Examples
    --------
    ### Pretty printing the `config` defaultdictionary:

    #### JINJA USAGE
    ```jinja2
    {{ config | pretty_defaultdict | safe }}
    ```

    The above jinja is roughly equivalent to something like this in pure python:

    ```python
    from ezcv.core import get_site_config

    config = get_site_config()

    print(pretty_defaultdict(config)) # Prints config dict in pretty form
    ```
    """
    logging.debug(f'[ezcv pretty_defaultdict({ugly_dict})]: Pretty printing defaultdict')

    return pformat(dict(ugly_dict)).replace("\n", "<br>").replace("{", "{<br>").replace("}", "<br>    }")



"""The module containing all cli functionality of ezcv including:
    - Initializing sites
    - Generating temporary preview
    - Getting lists of themes and/or copying themes

Functions
---------
init():
    Initializes an ezcv site

preview():
    Creates a temporary folder of the site's files and then previews it in browser

section():
    Creates a new section, or prints details about a section if it already exists

theme():
    Used to get information about the available themes and/or copy a theme folder

main():
    The primary entrypoint for the ezcv cli

Examples
--------
#### Create a new site
```
from ezcv.cli import init

theme = "freelancer"
name = "John Doe"

init(theme, name)
```

#### Preview a site that is in the cwd
```
from ezcv.cli import preview

preview()
```

#### Copy the aerial theme
```
from ezcv.cli import theme

theme(copy_theme = True, theme = "aerial")
```

#### Print a list of available themes
```
from ezcv.cli import theme

theme(list_themes = True)
```
"""

# Standard Lib Dependencies
from collections import defaultdict
import os                   # Used for path validation
import string               # Used for data validation
import shutil               # Used for file/folder copying and removal
import logging              # Used to log information for internal testing
import datetime             # Used for date formatting and date validation
from glob import glob       # Used to glob filepaths (patternmatch filepaths)
from sys import argv, exit  # Used to get length of CLI args and exit cleanly

## internal dependencies
from ezcv import __version__ as version
from ezcv.core import generate_site, get_site_config
from ezcv.themes import THEMES_FOLDER, generate_theme_metadata, get_remote_themes, get_theme_metadata, locate_theme_directory, setup_remote_theme
from ezcv.autoreload import start_server

# Third party dependencies
import yaml                      # Used to read and write yaml files
from colored import fg           # Used to highlight output with colors
from docopt import docopt        # Used to complete argument parsing for the cli
from PIL import Image            # Used to optimize and minify image files
from css_html_js_minify import * # Used to optimize and minify html/css/js files

usage = """Usage:
    ezcv [-h] [-v] [-p]
    ezcv build [-d OUTPUT_DIR] [-o]
    ezcv init [<name>] [<theme>] [-f]
    ezcv theme [-l] [-c] [-m] [<theme>]
    ezcv section <SECTION_NAME> [-t=<type>]


Options:
-h, --help            show this help message and exit
-v, --version         show program's version number and exit
-l, --list            list the possible themes
-c, --copy            copy the provided theme, or defined site theme
-p, --preview         preview the current state of the site
-o, --optimize        Optimize output files (takes longer to run)
-f, --flask           Generate Flask routes and requirements.txt
-d OUTPUT_DIR, --dir OUTPUT_DIR The folder name to export the site to
-m, --metadata        Generate metadata for the theme
-t=<type>, --type=<type> The type of section to generate [default: markdown]
"""

def init(theme_name="dimension", name="John Doe", flask:bool = False):
    """Initializes an ezcv site

    Parameters
    ----------
    theme_name : (str, optional)
        The theme_name to use in the config, by default "dimension"

    name : (str, optional)
        The name to use in the config, by default "John Doe"
    """
    if os.path.exists(os.path.abspath(name)): # If folder already exists at ./<name>
        logging.error("[ezcv cli.init()] The provided name already exists")
        print(f"[ezcv cli.init()] The provided name {name} already exists at {os.path.abspath(name)}")
        return
    logging.debug(f"[ezcv cli.init()] Initializing site with theme {theme_name} and name {name}")
    print(f"Generating site at {os.path.abspath(name)}")

    logging.debug(f"[ezcv cli.init()] Copying example site from {os.path.join(os.path.dirname(__file__), 'example_site')} to {os.path.abspath(name)}")
    shutil.copytree(os.path.join(os.path.dirname(__file__), "example_site"), os.path.abspath(name))

    # Generate initial config.yml file
    logging.debug("[ezcv cli.init()] generating config file")

    ## Get theme info 
    site_context= {"config": {"name": name, "theme": theme_name, "remotes": get_remote_themes()}}
    theme_folder = locate_theme_directory(theme_name, site_context)
    theme_metadata = get_theme_metadata(theme_folder)
    with open(os.path.join(name, "config.yml"), "w+") as config_file:
        config_file_contents = f"# See https://ezcv.readthedocs.io for documentation\nname: {name}\ntheme: {theme_name}\nresume: false\n"
        if theme_metadata["required_config"]: # Parse required config from theme if available
            if type(theme_metadata["required_config"]) == dict:
                for value in theme_metadata["required_config"]:
                    if value in ("name", "theme"):
                        continue
                    if not theme_metadata["required_config"][value].get("type", False):
                        theme_metadata["required_config"][value]["type"] = "str"
                    if not theme_metadata["required_config"][value].get("description", False):
                        theme_metadata["required_config"][value]["description"] = ""
                    else: # If description exists
                        theme_metadata["required_config"][value]["description"] = "# " + str(theme_metadata["required_config"][value]["description"])
                    if not theme_metadata["required_config"][value].get("default", False):
                        if theme_metadata["required_config"][value]["type"] == "str":
                            theme_metadata["required_config"][value]["default"] = f'"{str(value)}"'
                        elif theme_metadata["required_config"][value]["type"] == "bool":
                            theme_metadata["required_config"][value]["default"] = False
                        elif theme_metadata["required_config"][value]["type"] == "int":
                            theme_metadata["required_config"][value]["default"] = 0
                        elif theme_metadata["required_config"][value]["type"] == "datetime":
                            theme_metadata["required_config"][value]["default"] = f"{datetime.datetime.now().year}-{datetime.datetime.now().month}-{datetime.datetime.now().day}"
                    config_file_contents += f"{value}: {theme_metadata['required_config'][value]['default']} {theme_metadata['required_config'][value]['description']}\n"
            elif type(theme_metadata["required_config"]) == list:
                for value in theme_metadata["required_config"]:
                    if value in ("name", "theme"):
                        continue
                    config_file_contents += f"{value}: \"{str(value)}\"\n"
        config_file.write(config_file_contents)

    if theme_name != "dimension":
        logging.debug(f"[ezcv cli.init()] Not using dimension theme, copying theme {theme_name}")

        # Check if theme is remote theme, and download it if it is
        remote_themes = get_remote_themes()
        if remote_themes.get(theme_name, False):
            original_directory = os.path.abspath(os.getcwd()) # Store CWD
            os.chdir(os.path.abspath(name))                   # Go into new site folder
            setup_remote_theme(theme_name, remote_themes[theme_name])   # Download theme 
            os.chdir(original_directory)                      # Navigate back to original cwd

    if flask:
        logging.debug("[ezcv cli.init()] Finalizing Flask setup")
        os.remove(os.path.join(name, "standard-README.md"))
        os.rename(os.path.join(name, "flask-README.md"), os.path.join(name, "README.md"))
    else:
        logging.debug("[ezcv cli.init()] Finalizing Standard setup")
        os.remove(os.path.join(name, "routes.py"))
        os.remove(os.path.join(name, "requirements.txt"))
        os.remove(os.path.join(name, "flask-README.md"))
        os.rename(os.path.join(name, "standard-README.md"), os.path.join(name, "README.md"))

    print(f"Site generated and is available at {os.path.abspath(name)}")


def preview():
    """Creates a temporary folder of the site's files and then previews it in browser"""
    logging.debug("[ezcv cli.preview()] Generating temporary site preview")
    start_server()


def theme(list_themes: bool = False, copy_theme:bool = False, theme_name:str = "", metadata:bool = False):
    """Used to get information about the available themes and/or copy a theme folder

    Parameters
    ----------
    list_themes : bool, optional
        Whether or not to list the available themes, by default False

    copy_theme : bool, optional
        Whether or not to copy provided theme, by default False

    theme_name : str, optional
        The theme to copy, by default "" (which will copy the dimension theme)
    
    metadata : bool, optional
        Whether or not to generate metadata for the theme, by default False
    """
    logging.debug(f"[ezcv cli.theme({list_themes=}, {copy_theme=}, {theme_name=}, {metadata=})] Calling theme command")
    if not theme_name:
        logging.debug("[ezcv cli.theme()] No theme provided, using dimension theme")
        theme_name = "dimension"

    if metadata:
        if os.path.exists(os.path.join(THEMES_FOLDER, theme_name)): # If the theme exists in the themes folder
            try: # Try to copy the theme to ./<theme>
                logging.debug(f"[ezcv cli.theme()] Copying theme {theme_name} to ./{theme_name}")
                shutil.copytree(os.path.join(THEMES_FOLDER, theme_name), theme_name)
            except FileExistsError: # If a folder exists at ./<theme> remove and then re-copy
                shutil.rmtree(theme_name)
                shutil.copytree(os.path.join(THEMES_FOLDER, theme_name), theme_name)
            print(f"Copied {os.path.join(THEMES_FOLDER, theme_name)} to .{os.sep}{theme_name}")
            print(f"Generating/updating theme metadata for {theme_name}")
            logging.debug("[ezcv cli.theme()] Metadata specified, generating metadata file")
            metadata_path = os.path.abspath(f".{os.sep}{theme_name}{os.sep}metadata.yml")
            data = generate_theme_metadata(os.path.abspath(f".{os.sep}{theme_name}"), force=True)
            if not data["created"]:
                data["created"] = f"{datetime.datetime.now().year}-{datetime.datetime.now().month}-{datetime.datetime.now().day}"
            if not data["ezcv_version"]:
                data["ezcv_version"] = version
            data["updated"] = f"{datetime.datetime.now().year}-{datetime.datetime.now().month}-{datetime.datetime.now().day}"
            with open(metadata_path, "w+") as metadata_file:
                yaml.dump(dict(data), metadata_file)
        else: # Theme could not be found
            print(f"Theme {theme_name} not found and was unable to be copied")

    if copy_theme and not metadata:
        if os.path.exists(os.path.join(THEMES_FOLDER, theme_name)): # If the theme exists in the themes folder
            try: # Try to copy the theme to ./<theme>
                logging.debug(f"[ezcv cli.theme()] Copying theme {theme_name} to ./{theme_name}")
                shutil.copytree(os.path.join(THEMES_FOLDER, theme_name), theme_name)
            except FileExistsError: # If a folder exists at ./<theme> remove and then re-copy
                shutil.rmtree(theme_name)
                shutil.copytree(os.path.join(THEMES_FOLDER, theme_name), theme_name)
            print(f"Copied {os.path.join(THEMES_FOLDER, theme_name)} to .{os.sep}{theme_name}")
            print(f"Generating/updating theme metadata for {theme_name}")
            metadata_path = os.path.abspath(f".{os.sep}{theme_name}{os.sep}metadata.yml")
            data = generate_theme_metadata(os.path.abspath(f".{os.sep}{theme_name}"))
            if not data["created"]:
                data["created"] = f"{datetime.datetime.now().year}-{datetime.datetime.now().month}-{datetime.datetime.now().day}"
            if not data["ezcv_version"]:
                data["ezcv_version"] = version
            data["updated"] = f"{datetime.datetime.now().year}-{datetime.datetime.now().month}-{datetime.datetime.now().day}"
            with open(metadata_path, "w+") as metadata_file:
                yaml.dump(dict(data), metadata_file)
            
        else: # Theme could not be found
            print(f"Theme {theme_name} not found and was unable to be copied")

    if list_themes:
        # Get local themes
        logging.debug("[ezcv cli.theme()] Listing local themes")
        print(f"\nAvailable local themes\n{'='*22}")
        for current_theme in os.listdir(THEMES_FOLDER):
            if not current_theme == "remotes.yml":
                print(f"  - {current_theme}")

        # Get remote themes
        logging.debug("[ezcv cli.theme()] Getting remote themes")
        print(f"\nAvailable remote themes\n{'='*23}")
        with open(os.path.join(THEMES_FOLDER, "remotes.yml")) as remote_themes:
            for current_theme in yaml.safe_load(remote_themes):
                print(f"  - {current_theme}")
        print() # empty newline after list



def section(section_name:str, section_type:str):
    """Creates a new section, or prints details about a section if it already exists

    Parameters
    ----------
    section_name : str
        The name you want to give to the section, used to generate folder and template filename

    section_type : str
        The type of section you want to create, i.e. 'markdown'

    Notes
    -----
    - Needs to be run from the main folder of a site
    """    
    logging.debug(f"[ezcv cli.section({section_name=}, {section_type=})] Calling section command")

    # Sanatizing section name so resulting folder name is valid
    logging.debug(f"[ezcv cli.section()] Sanatizing {section_name=}")
    legal_path_characters = string.ascii_letters + string.digits+ " ()[]_-" # Allowed characters in file path
    section_name = section_name.replace(" ", "_")
    section_name = ''.join(current_character for current_character in section_name if current_character in legal_path_characters).lower()
    logging.debug(f"[ezcv cli.section()] Sanatized {section_name=}")

    if os.path.exists("config.yml"):
        config_path = "config.yml"
    else:
        print(f"{fg(1)}You are not in a project root folder, please run from folder with config.yml{fg(15)}\n")
        return 

    logging.debug("[ezcv cli.section()] Getting site config")
    config = get_site_config(config_file_path=config_path)
    if not config["theme"]: # Set to default theme if no theme is set
        config["theme"] = "dimension"

    logging.debug("[ezcv cli.section()] Getting theme metadata")
    theme_path = locate_theme_directory(config["theme"], {"config": config})
    theme_name = config["theme"]
    if os.path.exists(config["theme"]): # Theme is at cwd i.e. ./aerial
        theme_path = theme_name
    elif theme_path == os.path.join(THEMES_FOLDER, theme_name): # Theme is in package theme folder i.e. THEME_FOLDER/aerial
        try: # Try to copy the theme to ./<theme>
            print(f"[ezcv cli.section()] Copying theme {theme_name} to ./{theme_name}")
            logging.debug(f"[ezcv cli.section()] Copying theme {theme_name} to ./{theme_name}")
            shutil.copytree(os.path.join(THEMES_FOLDER, theme_name), config["theme"])
        except FileExistsError: # If a folder exists at ./<theme> remove and then re-copy
            shutil.rmtree(theme_name)
            shutil.copytree(os.path.join(THEMES_FOLDER, theme_name), theme_name)
        print(f"Copied {os.path.join(THEMES_FOLDER, theme_name)} to .{os.sep}{theme_name}")
        theme_path = theme_name
    else:
        print(f"{fg(1)}Could not find theme at any of the possible locations\n\t{config['theme']}\n\t{os.path.join('..', config['theme'])}\n\t{os.path.join(THEMES_FOLDER, config['theme'])} {fg(15)}\n")
        return 

    logging.debug("[ezcv cli.section()] Getting content directory")
    if os.path.exists("content"):
        content_path = "content"
    elif os.getcwd().split(os.sep)[-1] == "content": # Inside the current content folder
        content_path = os.getcwd()
    else:
        return
    
    logging.debug("[ezcv cli.section()] Getting theme metadata")
    theme_metadata = generate_theme_metadata(theme_path)
    if theme_metadata["sections"]:
        if theme_metadata["sections"].get(section_name, False):
            if theme_metadata["sections"][section_name].get("fields", False):
                fields_text = "\n\t• Fields: "
                for field in theme_metadata["sections"][section_name]["fields"]:
                    if type(theme_metadata["sections"][section_name]["fields"][field]) == str :
                        fields_text += f"\n\t\t • {field}: {theme_metadata['sections'][section_name]['fields'][field]} (optional)"
                    elif type(theme_metadata["sections"][section_name]["fields"][field]) == dict or type(theme_metadata["sections"][section_name]["fields"][field]) == defaultdict:
                        fields_text += f"\n\t\t • {field}: {theme_metadata['sections'][section_name]['fields'][field]['type']} ({'required' if theme_metadata['sections'][section_name]['fields'][field]['required'] else 'optional'})"
            else: # No fields specified
                fields_text = ""
            if theme_metadata["sections"][section_name].get("type", False) == "blog":
                print(f"""Section {section_name} details:
    • Type: Blog
    • Feed: {theme_metadata['sections'][section_name].get('feed', False)}
    • Single: {theme_metadata['sections'][section_name].get('single', False)}
    • Overview: {theme_metadata['sections'][section_name].get('overview', False)}{fields_text}
    """)
                return
            elif theme_metadata["sections"][section_name].get("type", False) == "gallery":
                print(f"Section {section_name} details:\n\t• Type: Gallery")
                return
            else:
                print(f"Section {section_name} details:\n\t• Type: Markdown{fields_text}")
            return
    if section_type.startswith("m"):
        # The content for the template in the generated section
        default_section_page_templte = f"""\n{{% for page in {section_name} %}} <!--Lets you iterate through each page -->
            {{{{ page[0] }}}} <!--Metadata access -->
            {{{{ page[1] | safe }}}} <!--content access -->
{{% endfor %}}
\n"""
    elif section_type.startswith("b"):
        section_type = section_type.replace("b", "").replace("blog","")
        # Parse which sections to include ([s]ingle.jinja, [f]eed.jinja, [o]verview.jinja)
        default_section_page_templte = defaultdict(lambda: False)
        if not section_type:
            section_type = "sfo" # If no additional type is specified, default to sfo
        if "s" in section_type:
            default_section_page_templte["single"] = f"""<!DOCTYPE html>\n<html>\n\t<body>
\t\t<h1>{{{{ content[0]['title'] }}}}</h1>
\t\t<h2>Created: {{{{ content[0]['created'] }}}} <br>Updated: {{{{ content[0]['updated'] }}}}</h1>
\t\t{{{{ content[1] | safe }}}}
\n\t</body>\n</html>"""
        if "f" in section_type:
            default_section_page_templte["feed"] = f"""{{% for post in {section_name} %}}
\t\t<h4> {{{{ post[0]["title"] }}}}</h4>
\t\t<!-- Metadata -->
\t\t<p>Created: {{{{ post[0]["created"] }}}}</p>
\t\t<p>updated:{{{{ post[0]["updated"] }}}}</p>
\t\t<a href='{{{{ post[0]["title"] }}}}'>{{{{ post[0]["title"] }}}}</a>
\t\t<!-- Content -->
\t\t<p>{{{{ post[1] | safe}}}}</p>
{{% endfor %}}"""
        if "o" in section_type:
            if "f" in section_type:
                default_section_page_templte["overview"] = f"""<!DOCTYPE html>\n<html>\n\t<body>
{{{{ {section_name}_html | safe}}}}
\n\t</body>\n</html>"""
            else:
                default_section_page_templte["overview"] = f"""<!DOCTYPE html>\n<html>\n\t<body>
{{% for post in {section_name} %}}
\t\t<h4> {{{{ post[0]["title"] }}}}</h4>
\t\t<!-- Metadata -->
\t\t<p>Created: {{{{ post[0]["created"] }}}}</p>
\t\t<p>updated:{{{{ post[0]["updated"] }}}}</p>
\t\t<!-- Content -->
\t\t<p>{{{{ post[1] | safe}}}}</p>
{{% endfor %}}
\n\t</body>\n</html>"""
    elif section_type.startswith("g"):
        # The content for the template in the generated section
        default_section_page_templte = f"""\n{{% for image in {section_name} %}} <!--Lets you iterate through each image -->
            {{{{ image[0] }}}} <!--Metadata -->
            <img src="{{{{ image[0]['file_path'] }}}}" alt="{{{{ image[0]['file_path'].split()[-1] }}}}"> <!--Image -->
{{% endfor %}}
\n"""
    if not os.path.exists(os.path.join(theme_path, "sections")):
        os.mkdir(os.path.join(theme_path, "sections"))
    # Begin creating content folder and template file(s)
    if type(default_section_page_templte) == str:
        if not os.path.exists(os.path.join(content_path, section_name)): # If the content folder doesn't already exist
            if not os.path.exists(os.path.join(theme_path, "sections", f"{section_name}.jinja")) and not os.path.exists(os.path.join(theme_path, "sections", f"{section_name}.html")): # If jinja theme doesn't already exist
                logging.debug("[ezcv cli.new_section()] Creating section folder")
                if not os.path.exists(os.path.join(content_path, section_name)):
                    os.mkdir(os.path.join(content_path, section_name))
                logging.debug("[ezcv cli.new_section()] Creating section theme file")
                if not os.path.exists(os.path.join(theme_path, "sections")):
                    os.mkdir(os.path.join(theme_path, "sections"))
                with open(os.path.join(theme_path, "sections", f"{section_name}.jinja"), 'w+') as section_file:
                    section_file.write(default_section_page_templte)
            else: # Theme file already existed
                print(f"{fg(1)}Could not create path, path already exists at either: \n\t{os.path.join(theme_path, 'sections', f'{section_name}.jinja')}\n\tor\n\t{os.path.join(theme_path, 'sections', f'{section_name}.jinja')}\n{fg(15)}")
                return 
        else: # Content folder already existed
            print(f"{fg(1)}Could not create path, path already exists at {os.path.join(content_path, section_name)}\n{fg(15)}")
            return 
        if not theme_metadata["sections"]:
            theme_metadata["sections"] = {}
        theme_metadata["sections"][section_name] = {"type": "markdown" if "m" in section_type else "gallery"}
        print(f"Section successfully created\n\nTheme file created at:\n\t{os.path.join(theme_path, 'sections', f'{section_name}.jinja')}\nContent folder created at:\n\t{os.path.join(content_path, section_name)}")
    elif type(default_section_page_templte) == defaultdict:
        os.mkdir(os.path.join(theme_path, "sections", section_name))
        if not os.path.exists(os.path.join(content_path, section_name)):
            os.mkdir(os.path.join(content_path, section_name))
        if default_section_page_templte["single"]:
            with open(os.path.join(theme_path, "sections", section_name, "single.jinja"), 'w+') as section_file:
                section_file.write(default_section_page_templte["single"])
        if default_section_page_templte["feed"]:
            with open(os.path.join(theme_path, "sections", section_name, "feed.jinja"), 'w+') as section_file:
                section_file.write(default_section_page_templte["feed"])
        if default_section_page_templte["overview"]:
            with open(os.path.join(theme_path, "sections", section_name, "overview.jinja"), 'w+') as section_file:
                section_file.write(default_section_page_templte["overview"])
        if not theme_metadata["sections"]:
            theme_metadata["sections"] = {}
        theme_metadata["sections"][section_name] = {"fields":{"title": {"required": True, "type":"str"},"created": {"required": True, "type":"str"},"updated": {"required": True, "type":"str"}, },"type": "blog"}
        theme_metadata["sections"][section_name]["single"] = bool(default_section_page_templte["single"])
        theme_metadata["sections"][section_name]["overview"] = bool(default_section_page_templte["overview"])
        theme_metadata["sections"][section_name]["feed"] = bool(default_section_page_templte["feed"])
        print(f"Section successfully created\n\nTheme file(s) created at:\n\t{os.path.join(theme_path, 'sections', f'{section_name}')}(remember to add your CSS)\nContent folder created at:\n\t{os.path.join(content_path, section_name)}")
    with open(os.path.join(theme_path, "metadata.yml"), 'w+') as metadata_file:
        yaml.dump(dict(theme_metadata), metadata_file)

def optimize(directory:str = "site"):
    """Goes through and minifies html, css, js and image files in directory

    Notes
    -----
    - This assumes you are following the standard template layout: https://ezcv.readthedocs.io/en/latest/theme-development/#folder-layout
    - Only image extensions supported are:

        - .jpg
        - .png
        - .jpeg

    Parameters
    ----------
    directory : str
        The directory you want to minify all files from
    """
    logging.debug(f"[ezcv cli.optimize({directory=})] Optimizing files")
    # Minify html/css/js
    html = glob(f'{directory}{os.sep}*.html')
    css = glob(f'{os.path.join(os.path.join(directory, "css"))}{os.sep}*.css')
    js = glob(f'{os.path.join(os.path.join(directory, "js"))}{os.sep}*.js')

    for file in html:
        logging.debug(f"[ezcv cli.optimize()] Processing html file: {file}")
        process_single_html_file(file, overwrite=True)
    
    for file in css:
        logging.debug(f"[ezcv cli.optimize()] Processing css file: {file}")
        process_single_css_file(file, overwrite=True)
    
    for file in js:
        logging.debug(f"[ezcv cli.optimize()] Processing js file: {file}")
        process_single_js_file(file, overwrite=True)

    # Find and process images
    png = glob(f'{os.path.join(os.path.join(directory, "images"))}{os.sep}*.png')
    jpg = glob(f'{os.path.join(os.path.join(directory, "images"))}{os.sep}*.jpg')
    jpeg = glob(f'{os.path.join(os.path.join(directory, "images"))}{os.sep}*.jpeg')

    for extension in [png, jpg, jpeg]:
        if extension: # if list is not empty
            for image in extension:
                logging.debug(f"[ezcv cli.optimize()] Processing image file: {image}")
                pil_object = Image.open(image)
                pil_object.save(image, optimize=True, quality=85)

def main():
    """The primary entrypoint for the ezcv cli"""
    args = docopt(usage, version=version)
    logging.debug(f"[ezcv cli.main()] Arguments: {args}")

    if len(argv) == 1: # Print usage if no arguments are given
        print("\n", usage)
        exit()

    if args["--preview"] and not args["build"]:
        preview()

    elif args["init"]:
        if args["<theme>"] and args["<name>"]: # Both a theme and name are specified
            init(args["<theme>"], args["<name>"], flask=args["--flask"])
        elif args["<name>"]: # Only a name is specified
            init(name = args["<name>"], flask=args["--flask"])
        else: # No values are specified
            init(flask=args["--flask"])
        exit()

    elif args["build"]:
        if not args["--dir"]:
            generate_site()
        else:
            generate_site(args["--dir"])

        if args["--optimize"]:
            if args["--dir"]:
                optimize(args["--dir"])
            else:
                optimize()
        exit()

    elif args["theme"]:
        if args["--metadata"]:
            if not args["<theme>"]:
                args["<theme>"] = get_site_config()["theme"]
            theme(args["--list"], args["--copy"], args["<theme>"], metadata=True)
            exit()
        elif args["<theme>"]:
            theme(args["--list"], args["--copy"], args["<theme>"])
            exit()
        elif args["--copy"]: # If copy is flagged, but no theme is provided
            if os.path.exists("config.yml"):
                theme(args["--list"], args["--copy"], get_site_config()["theme"])
            else: # If no theme, or config.yml file is present
                theme(args["--list"], args["--copy"], "freelancer")
            exit()
        elif args["--list"]:
            theme(args["--list"])
            exit()
        else: # If theme argument is called with no other flags
            print("\n", usage)
            exit()
    
    elif args["section"]:
        if not args["--type"]:
            args["--type"] = "markdown"
        section(args["<SECTION_NAME>"], args["--type"])
        exit()

    elif args["--preview"]: # If preview flag is specified with no other flags
        preview()
        exit()

    else: # No top level argument is provided
        print("\n", usage)
        exit()

if __name__ == "__main__": # For testing
    main()



"""The module containing all primary functionality of ezcv including:

- Content parsing
- HTML generation
- Site exporting

Functions
---------
generate_site():
    The primary entrypoint to generating a site

get_site_config() -> defaultdict:
    Gets the site config from provided file path and returns defaultdict of values


Module Variables
----------------
SECTIONS_LIST (list[str]):
    The list of the first party supported sections

Examples
--------
#### Generating a site using all settings defined in "config.yml"
```
from ezcv.core import generate_site

generate_site()
```

#### Generating a site overriding the theme in "config.yml", output directory and specifying to show a preview of the site
```
from ezcv.core import generate_site

generate_site(output_folder="my_site", theme = "aerial", preview = True)
```
"""

# Standard Lib Dependencies
import os                           # Used for path validation
import shutil                       # Used for file/folder copying and removal
import logging                      # Used to log information for internal testing
from collections import defaultdict # Used to instatiate dictionaries with default arguments on unspecified keys
from typing import Callable, Union  # Used to add additional typehints to help with documentation and usage on functions

# Internal Dependencies
from ezcv.themes import *
from ezcv.content import *
from ezcv.filters import inject_filters

# Third Party Dependencies
import yaml                         # Used for config file parsing
import jinja2                       # used as middlewear for generating templates
from tqdm import tqdm               # Used to generate progress bars during iteration

# The global list of currently supported first party sections
SECTIONS_LIST = ["projects", "education", "work_experience", "volunteering_experience", "gallery", "blog"]

def get_site_config(config_file_path:str = "config.yml", remotes_file_path:str = os.path.join(THEMES_FOLDER, "remotes.yml")) -> defaultdict:
    """Gets the site config from provided file path and returns defaultdict of values

    Parameters
    ----------
    config_file_path : str, optional
        The path to the config file, by default "config.yml"

    Returns
    -------
    defaultdict:
        The configuration, if any key is not present it defaults to False
    """
    logging.debug(f"[ezcv get_site_config({config_file_path}, {remotes_file_path})]: Loading config file")
    if not os.path.exists(config_file_path):
        raise FileNotFoundError(f"Config file at {config_file_path} was not found")

    with open(config_file_path, "r") as config_file:
        config = yaml.safe_load(config_file)

    logging.debug(f"[ezcv get_site_config({config_file_path}, {remotes_file_path})]: Loading remotes file")
    config["remotes"] = get_remote_themes(remotes_file_path)

    # Convert config dict to defaultdict so that all empty values are False instead of giving KeyNotFoundError
    default_dict_config = defaultdict(lambda: False, config)

    return default_dict_config


def _render_section(section_name:str, site_context:dict, environment:jinja2.Environment, blog:bool=False) -> str:
    """Renders the particular section provided using the environment provided

    Parameters
    ----------
    section_name : (str)
        The name of the section to render i.e. projects, education, work_experience etc.

    site_context: (dict)
        The dictionary containing the site's context

    environment : (jinja2.Environment)
        The jinja environment pre-loaded with the themes and filters

    blog : (bool, optional)
        Wheter or not the section is a blog, by default False

    Returns
    -------
    str:
        The rendered template of the section
    """
    logging.debug(f"[ezcv _render_section({section_name}, {site_context}, {environment})]: Begin rendering section")
    try:
        contents = site_context["sections"][section_name]
    except KeyError:
        print(f"Could not find content for section '{section_name}', skipping")
        if blog:
            return "", "", ""
        else:
            return ""

    logging.debug(f"[ezcv _render_section({section_name}, {site_context}, {environment})]: Rendering sections")
    if not blog: # Rendering sections that are not blog sections
    # If a section template exists set it to the path, else False i.e. if <theme folder>/sections/<section name>.jinja exists set it to that
        section_template_file = f"sections/{section_name}.jinja"
        if len(contents) > 0: # If there is any markdown content
            try:
                logging.debug("[ezcv _render_section()] Found markdown section")
                theme = environment.get_template(section_template_file)
            except jinja2.TemplateNotFound: # If current section is not supported
                print(f"Section {section_name} template is not available")
                return ""
            return theme.render({section_name:contents, "config": site_context["config"]})
        else:
            return ""
    else: # Rendering blog sections
        logging.debug("[ezcv _render_section()] Found blog section")
        overview_file = f"sections/{section_name}/overview.jinja"
        feed_file = f"sections/{section_name}/feed.jinja"
        single_file = f"sections/{section_name}/single.jinja"
        theme_folder = environment.loader.searchpath[0]
        overview_file_path = os.path.join(theme_folder, overview_file)
        feed_file_path = os.path.join(theme_folder, feed_file)
        single_file_path = os.path.join(theme_folder, single_file)
        if not overview_file_path:
            logging.debug("[ezcv _render_section()] No overview template file found")
            overview_file = ""
        if not feed_file_path:
            logging.debug("[ezcv _render_section()] No feed template file found")
            html = ""
        elif feed_file_path:
            if len(contents) > 0: # If there is any markdown content
                logging.debug("[ezcv _render_section()] Rendering feed file")
                html = environment.get_template(feed_file).render({section_name:contents, "config": site_context["config"]})
            else:
                html = ""
        if not single_file_path:
            logging.debug("[ezcv _render_section()] No single template file found")
            single_file = ""
        return single_file, overview_file, html


def _render_page(page:str, site_context:dict, environment:jinja2.Environment) -> str:
    """Renders the page provided from the specified theme

    Parameters
    ----------
    page : (str)
        The filename inside the theme folder to render i.e. 'index.jinja'

    site_context : (dict)
        A dictionary containing the config values, and all sections html

    environment : (jinja2.Environment)
        The jinja environment pre-loaded with the themes and filters

    Returns
    -------
    str:
        The rendered html of the page
    """
    logging.debug(f"[ezcv _render_page({page}, {site_context}, {environment})]: Begin rendering page")
    inject_filters(environment) # Add in custom filters
    # Render template and return contents
    theme = environment.get_template(page)
    return theme.render(site_context)


def _export(site_context:dict, theme_folder:str, environment:jinja2.Environment, output_folder:str = "site",  pages:list=None ):
    """Generates all the site html from pages specified and outputs them to the output folder

    Parameters
    ----------
    site_context : (dict)
        The site context containing all sections html and dict + config dict

    theme_folder : (str)
        The absolute path to the folder for the theme to use

    environment : (jinja2.Environment)
        The jinja environment pre-loaded with the themes and filters

    output_folder : (str, optional)
        The folder to output the HTML files to, by default "site"

    pages : (list, optional)
        The list of pages to use, by default None which gets set to ["index.jinja"]

    Raises
    ------
    FileNotFoundError
        If the provided theme folder does not exist
    """
    if pages is None:
        pages = ["index.jinja"]
    if not os.path.exists(theme_folder): # Error out if provided theme folder does not exist
        raise FileNotFoundError(f"The provided theme folder does not exist: {theme_folder}")

    # Copy source files
    try:
        shutil.copytree(theme_folder, output_folder, ignore=shutil.ignore_patterns("*.jinja"))

    except FileExistsError:
        shutil.rmtree(output_folder)
        shutil.copytree(theme_folder, output_folder, ignore=shutil.ignore_patterns("*.jinja"))
    
    # Copy images
    output_image_dir = os.path.join(output_folder, "images")
    if os.path.exists("images"):
        if not os.path.exists(output_image_dir): # Create output_folder/images if it's not present
            os.mkdir(output_image_dir)
        for file in os.listdir("images"): # Copy file from source images folder to output image directory
            # TODO: Add catch for if image already exists
            shutil.copyfile(os.path.join("images", file), os.path.join(output_image_dir, file))

    # Copy Gallery images
    output_fallery_image_dir = os.path.join(output_folder, "images", "gallery")
    if os.path.exists(os.path.join("content", "gallery")):
        if not os.path.exists(output_fallery_image_dir): # Create output_folder/images/gallery if it's not present
            os.mkdir(output_fallery_image_dir)
        for file in os.listdir(os.path.join("content", "gallery")): # Copy file from source images folder to output image directory
            # TODO: Add catch for if image already exists
            shutil.copyfile(os.path.join("content", "gallery", file), os.path.join(output_fallery_image_dir, file))

    # Iterate through top level pages and write to the output folder
    print("\nGenerating output html from theme")
    pages_iterator = tqdm(pages)
    pages_iterator.set_description_str("Generating top level pages")
    templates_list = environment.list_templates()
    for page in pages_iterator:  # Write new pages
        if type(page) == str: # Standard markdown sections
            try:
                html = _render_page(page, site_context, environment)
            except jinja2.UndefinedError as e:
                print(e)
                raise ValueError("A required configuration value is missing")
            if page.endswith(".jinja"):
                page = f"{page[:-6:]}.html"
            pages_iterator.set_description_str(f"Writing {page}")
            pages_iterator.refresh()
            with open(f"{output_folder}{os.sep}{page}", "w+") as outfile:
                outfile.write(html)
        elif type(page) == list: # Blog sections
            if len(page) == 2: # overview pages
                logging.debug(f"[ezcv _export()]: Rendering {page[0]} overview page")
                if not "sections/blog/overview.jinja" in templates_list:
                    print("[ezcv _export()]: No overview template found")
                    continue
                try:
                    html = _render_page(page[1], site_context, environment)
                except jinja2.UndefinedError as e:
                    print(e)
                    raise ValueError("A required configuration value is missing")
                if page[1].endswith(".jinja"):
                    page = f"{page[0]}.html"
                pages_iterator.set_description_str(f"Writing {page}")
                pages_iterator.refresh()
                with open(f"{output_folder}{os.sep}{page}", "w+") as outfile:
                    outfile.write(html)
            elif len(page) == 3: # Single pages
                logging.debug(f"[ezcv _export()]: Rendering {page[0]} single pages")
                if not "sections/blog/single.jinja" in templates_list:
                    print("[ezcv _export()]: No single blog post template found")
                    continue
                template_file = page[1]
                for content_file in page[2]:
                    if content_file[0]["title"]:
                        title = content_file[0]["title"]
                    else:
                        title = content_file[2].replace('.md', '')
                    if title == "index":
                        raise ValueError("The title of a blog post cannot be 'index'")
                    try:
                        single_page_context = {"config": site_context["config"], "content": [content_file[0], content_file[1]]}
                        html = _render_page(template_file, single_page_context, environment)
                    except jinja2.UndefinedError as e:
                        print(e)
                        raise ValueError("A required configuration value is missing")
                    if template_file.endswith(".jinja"):
                        page = f"{title}.html"
                    pages_iterator.set_description_str(f"Writing {page}")
                    pages_iterator.refresh()
                    with open(f"{output_folder}{os.sep}{page}", "w+") as outfile:
                        outfile.write(html)
    logging.debug("Cleaning up metadata.yml")
    if os.path.exists(os.path.join(output_folder, "metadata.yml")):
        os.remove(os.path.join(output_folder, "metadata.yml")) # Remove metadata file


def generate_site(output_folder:str="site", theme:str = "dimension", sections: list = None, config_file_path="config.yml", preview:bool = False, extra_filters:List[Callable] = None):
    """The primary entrypoint to generating a site

    Parameters
    ----------
    output_folder : (str, optional)
        The folder to output the site files to, by default "site"

    theme : (str, optional)
        The name of the theme to use, by default "dimension"

    sections : (list[str], optional)
        A list of the sections to include in export, by default []

    config_file_path : (str, optional)
        The path to the site's config yaml file, by default "config.yml"

    preview : (bool, optional)
        Changes generation to work for the autoreload server, by default False

    extra_filters : List[Callable], optional
        An optional set of method objects containing additional filter functions you want to use

    Notes
    -----
    - theme options are: 
        - aerial; https://html5up.net/aerial
        - base; included theme that can be used for debugging
        - creative; https://startbootstrap.com/theme/creative
        - cv; https://startbootstrap.com/theme/resume
        - dimension; https://html5up.net/dimension
        - ethereal; https://html5up.net/ethereal
        - freelancer; https://startbootstrap.com/theme/freelancer
        - Grayscale; https://startbootstrap.com/theme/grayscale
        - identity; https://html5up.net/identity
        - Lens; https://html5up.net/lens
        - Paradigm Shift; https://html5up.net/paradigm-shift
        - read_only; https://html5up.net/read-only
        - solid_state; https://html5up.net/solid-state
        - strata; https://html5up.net/strata
    - Available sections are:
        - Projects (projects)
        - Education (education)
        - Volunteering experience (volunteering_experience)
        - Work Experience (work_experience)
        - Gallery (gallery)
    - If sections is an empty list then the theme's section directory will be searched for themes

    Raises
    ------
    FileNotFoundError
        If the provided theme folder does not exist

    Examples
    --------
    Generating a site with all default settings
    ```
    from ezprez.core import generate_site

    generate_site()
    ```

    Generating a site that outputs to /resume and only generating the projects section
    ```
    from ezprez.core import generate_site

    generate_site(output_folder="my_site", sections=["projects"])
    ```
    """
    if sections is None:
        sections = []
    if extra_filters is None:
        extra_filters = []
    print(f"Exporting site to {output_folder}")
    pages = [] # Filled with a list of all the pages to render

    # The data passed to render all pages
    logging.debug("[ezcv] Initializing site context with config file")
    site_context:dict[str, Union[list, defaultdict, dict]] = {"config": get_site_config(config_file_path)}

    logging.debug("[ezcv] Getting ignore_exif_data config value")
    if site_context["config"]["ignore_exif_data"]:
        Image.ignore_exif_data = True

    # If no theme argument, and a theme is defined in the site config file
    logging.debug("[ezcv] Getting theme config value")
    if site_context["config"]["theme"] and theme == "dimension": 
        logging.debug(f"Using theme from site config file: {site_context['config']['theme']}")
        theme = site_context["config"]["theme"]

    # Find theme directory based on name, or download it if it's a remote theme
    logging.debug("[ezcv] Getting theme directory")
    theme_folder = locate_theme_directory(theme, site_context)
    logging.info(f"[ezcv] theme directory: {theme_folder}" )

    # Check required_config values
    if not os.path.exists(os.path.join(theme_folder, "metadata.yml")):
        new_metadata = dict(generate_theme_metadata(theme_folder))
        with open(os.path.join(theme_folder, "metadata.yml"), "w+") as outfile:
            yaml.dump(new_metadata, outfile)
    theme_metadata = get_theme_metadata(theme_folder)
    if theme_metadata["required_config"]:
        for value in theme_metadata["required_config"]:
            if not value in site_context["config"].keys():
                if not isinstance(theme_metadata["required_config"][value], dict):
                    theme_metadata["required_config"][value] = {"type": "str", "default": "", "description": ""}
                theme_metadata["required_config"][value]["default"] = theme_metadata["required_config"][value].get("default", "")
                theme_metadata["required_config"][value]["type"] = theme_metadata["required_config"][value].get("type", "str")
                theme_metadata["required_config"][value]["description"] = theme_metadata["required_config"][value].get("description", "")
                print(f"\n\x1b[31mThe theme requires the '{value}'configuration value \n\n\ttype: { theme_metadata['required_config'][value]['type'] } \n\tdescription: { theme_metadata['required_config'][value]['description'] }\n\n please add\n\n\x1b[37m\t {value}: <value> \n\n\x1b[31mto your config.yml file\x1b[37m")
                exit(1)

    # Initialize jinja loaders
    logging.debug("[ezcv] Initializing jinja2 loaders")
    theme_loader = jinja2.FileSystemLoader(theme_folder)
    environment = jinja2.Environment(loader=theme_loader, autoescape=True, trim_blocks=True) # Grab all files in theme_folder

    logging.debug("[ezcv] Injecting extra jinja filters into environment (if available)")
    environment = inject_filters(environment, extra_filters)

    # Initialize sections key in site context to empty dict
    site_context["sections"] = {}

    # Get a list of the section names, and section theme directories
    logging.debug("[ezcv] Getting info about sections content")
    sections = get_theme_section_directories(theme_folder, sections, preview) # TODO: Add support for blog sections
    sections_content_dirs = get_content_directories()
    logging.info(f"[ezcv] Found sections: {sections}\n[ezcv] Found section content directories: {sections_content_dirs}" )

    # Go through all section content files to get content (i.e. ./sections/education/*.md)
    logging.debug("[ezcv] Getting content for each section")
    for section in sections_content_dirs: 
        if not section.split(os.sep)[-1] == "blog": #TODO: make parametric
            # Get content to store in site_context["sections"][section]
            site_context["sections"][section.split(os.sep)[-1]] = get_section_content(section, site_context["config"]["examples"])
        else:
            # Get content to store in site_context["sections"][section]
            site_context["sections"][section.split(os.sep)[-1]] = get_section_content(section, site_context["config"]["examples"], blog=True)

    # Get a list of all the top level pages in the theme folder and add them to the pages list
    logging.debug("[ezcv] Getting list of top-level files (.jinja and .html)")
    for top_level_file in os.listdir(theme_folder):
        if top_level_file == "resume.jinja" and not site_context["config"]["resume"]: # Ignore resume.jinja if resume config var is False
            continue
        # TODO: add support for blog overview files
        if top_level_file.endswith(".jinja") or top_level_file.endswith(".html"):
            pages.append(top_level_file)
    
    # Go through each section, render the html and add it to the site context
    print("\nGenerating content from sections")
    logging.debug("[ezcv] Generating html from section content")
    sections_iterator = tqdm(sections)
    sections_iterator.set_description_str("Writing section content")
    for section in sections_iterator:
        if section == "blog": #TODO: Make parametric based on setup
            single_page, overview_page, feed_html = _render_section(section, site_context, environment, blog=True)
            if single_page or overview_page or feed_html:
                site_context[f"{section}_html"] = feed_html
                pages.append([section, overview_page])
                pages.append([section, single_page, site_context["sections"][section]])
        else:
            html = _render_section(section, site_context, environment)
            site_context[f"{section}_html"] = html

    # Generate and export all the pages of a site
    logging.debug("[ezcv] Generating html from pages")
    _export(site_context, theme_folder, environment, output_folder, pages)
